#!/usr/bin/python3
import os,tempfile
import time
import logging
import threading
from apt import Cache
import requests
import hashlib
import queue
import subprocess
import shutil
from gettext import gettext as _
from shutil import rmtree
from os.path import abspath
from gettext import gettext as _

from .UpdaterConfigParser import UpgradeConfig
from .enums import *
from .errors import *
# from Core.MyCache import MyCache

#accountsservice_0.6.55-0kylin11k3_0.6.55-0kylin12~20.04.6_arm64.debdelta
DELTAS_NAME = "%s_%s_%s_%s.debdelta"

# http://debdeltas.debian.net/debian-deltas +'/'+ os.path.dirname(deb_path)+'/'+base-files_11.1+deb11u6_11.1+deb11u7_amd64.debdelta
DELTAS_URL = "%s/%s/%s"
DEB_DIR = '/var/cache/apt/archives/'
HTTP_USER_AGENT = {'User-Agent':'Debian debdelta-upgrade','Range':'bytes=0-10239',}
TMPDIR = '/tmp'

DPKG_KEEPS_CONTROLS = ('conffiles','config','list','md5sums','postinst',
                                    'postrm','preinst','prerm','shlibs','templates')

class DeltaDownloadBackend():
    """
    Represent the (potentially partial) results of an unattended-upgrades
    run
    """
    def __init__(self,parent):
        self.cache = parent.cache
        self.configs_cover = parent.configs_cover

        # self.sources_config = UpgradeConfig(datadir = "/etc/debdelta/", name = "sources.conf")
        self.diversions = self._scan_diversions()

    def commit_download(self,upgrade_list,save_patch=DEB_DIR):
        available_deltas = []
        not_available_deltas = []
        no_delta = []

        # 合并开启线程
        patching_queue=queue.Queue()
        patching_thread=threading.Thread(target=self._thread_do_patch,args=(patching_queue,available_deltas,not_available_deltas,) )
        patching_thread.daemon=True
        patching_thread.start()

        # 建立下载ssesion
        download_session = requests.Session()
        download_session.headers.update(HTTP_USER_AGENT.copy())

        delta_uri = self.configs_cover.getWithDefault("main debian archive", "delta_uri",None)
        if delta_uri == None:
            return

        for pkg_name in upgrade_list:
            
            if pkg_name not in self.cache:
                logging.warning('Check: packages(%s) not in Cache',pkg_name)
                not_available_deltas.append(pkg_name)
                continue

            pkg_obj = self.cache[pkg_name]
            if not pkg_obj.is_installed or not pkg_obj.is_upgradable:
                logging.warning('Check: packages(%s) not upgradable or installed',pkg_name)
                not_available_deltas.append(pkg_name)
                continue
            
            # "pool/main/a/accountsservice/accountsservice_0.6.55-0kylin12~20.04.6_arm64.deb"
            deb_path = getattr(pkg_obj.candidate, "filename", '')
            if deb_path == '':
                logging.warning('Check: candidate packages(%s) filename is empty: %s',pkg_name)
                not_available_deltas.append(pkg_name)
                continue
            
            # accountsservice_0.6.55-0kylin12~20.04.6_arm64.deb
            new_debfile = deb_path.split('/')[-1]
            if os.path.exists(save_patch + new_debfile):
                logging.info('Check: Already downloaded: %s %s',pkg_name,new_debfile)
                available_deltas.append(pkg_name)
                continue
            
            # accountsservice_0.6.55-0kylin11k3_0.6.55-0kylin12~20.04.6_arm64.debdelta
            delta_name = DELTAS_NAME % (pkg_obj.shortname,pkg_obj.installed.version,\
                                                pkg_obj.candidate.version,pkg_obj.candidate.architecture)

            # 下载的url
            download_uri= DELTAS_URL % (delta_uri,os.path.dirname(deb_path),delta_name)

            # 保存的位置
            delta_path = save_patch + delta_name

            # FIXME:检查delta文件已经下载存在本地 就直接进行合成
            if os.path.exists(delta_path) or os.path.exists(save_patch+'partial/'+delta_name):
                logging.debug('Already here: %s',delta_path)
                patching_queue.put((pkg_obj,delta_path,save_patch + new_debfile))

            logging.debug("Package(%s) delta download url:",pkg_name,download_uri)
            success,status,data = self._download_10k_uri(download_session,download_uri)

            # 检查状态码 其他错误码就认为失败 直接返回
            if not success or data == None:
                logging.warning("Delta is not downloadable(%d) %s",status,download_uri)
                not_available_deltas.append(pkg_name)
                continue
            
            # 保存文件到本地
            self._save_to_local(data,delta_path)

            if success:
            # 放入开始合成
                patching_queue.put( (pkg_obj,delta_path,save_patch + new_debfile))
            else:
                pass
                # available_deltas.append( (l, p.name, uri, delta_path , newdeb, deb_uri, tempname, complete) )

        #关闭会话
        download_session.close()
        # 终止合并线程
        patching_queue.put(None)

        while patching_thread.is_alive():
            time.sleep(0.1)

        logging.info("Finished: Delta Download available_deltas:%r not_available_deltas:%r",available_deltas,not_available_deltas)

        return available_deltas,not_available_deltas

    def _thread_do_patch(self,que,available,not_available):

        logging.info('  Patching thread started. ')
        while True:
            task = que.get()
            if task == None:
                break
            (pkg_obj,delta_path,new_deb_path) = task

            logging.debug('  Now patching for: %s',pkg_obj.name)
            try:
                # 进行合成的地方 将旧包合成新包 _do_patch("/home/x/test.debdelta","/","/home/x/test.deb",[],cache["kylin-system-updater"])
                self._do_patch(pkg_obj,delta_path,new_deb_path,self.diversions)
                available.append(pkg_obj.name)
            except KeyboardInterrupt:
                return
            except DebDeltaError:
                not_available.append(pkg_obj.name)
            else:
                # 删除delta文件
                if os.path.exists(delta_path):
                    os.unlink(delta_path)
        logging.info('Patching thread ended...')

    # _download_10k_uri
    def _download_10k_uri(self,download_session,uri):
        data = None
        success = False
        status = 404
        try:
            req = download_session.get(uri,timeout = 1,stream = True)
        except requests.exceptions.ConnectionError as e:
            logging.warning(str(e))
            return success,status,data
        except requests.exceptions.RequestException as e:
            logging.warning(str(e))
            return success,status,data
        
        status = req.raw.status

        if status == 206 or status == req.raw.status:
            data = req.raw.read()
            success = True

        req.close()
        
        return success,status,data

    """
        delta = '/tmp/archives/base-files_11.1+deb11u6_11.1+deb11u7_amd64.debdelta'
        olddeb = '/'
        newdeb = '/tmp/archives/base-files_11.1%2bdeb11u7_amd64.deb'
        
    """
    def _do_patch(self,pkg_cache,debdelta_file,new_path,diversions):
        try:
            tmp_dir = self._make_temp_dir()

            # 转化为绝对路径        
            debdelta_file = abspath(debdelta_file)
            new_path = new_path
            olddeb = abspath('/')

            # 检查是不是delta文件的格式
            self._check_is_delta(debdelta_file)

            if not os.path.isdir(olddeb):
                # 检查是不是deb包
                self._check_deb(olddeb)

            # 从debdlta文件提取内容 执行 ar xvo /tmp/archives/base-files_11.1+deb11u6_11.1+deb11u7_amd64.debdelta
            # 解压到指定目录 /tmp/debdeltaxv0h1pvi/PATCH 下 内容 info 0 1 2 patch.sh
            ret,output,error  = self.run_system("ar xvo "+ debdelta_file,tmp_dir+'/PATCH')

            if ret:
                logging.error('Cannot extract from '+ debdelta_file)
                raise DebDeltaError(ERROR_UPDATE_DEFAULT_FAILED)

            # FIXME: 检查签名

            # 解压出patch.sh 文件
            self._info_patch_unzip(tmp_dir)

            if not os.path.isfile(tmp_dir+'PATCH/patch.sh'):
                logging.error("Error. File '%s' is not a delta file." % debdelta_file)
                raise DebDeltaError(ERROR_UPDATE_DEFAULT_FAILED)

            # 创建软链接 到/tmp/debdelta2i52er9m/ 构建差分包的目录
            os.symlink('/usr/lib/debdelta/minigzip', tmp_dir+'minigzip')
            os.symlink('/usr/lib/debdelta/minibzip2', tmp_dir+'minibzip2')

            # 主要进行读取info文件
            info = self._scan_delta_info(tmp_dir)
            params = self.info_2_db(info)

            # FIXME:  提前检查所有磁盘空间

            self._prepare_env_from_params(pkg_cache,params,tmp_dir,olddeb,diversions)

            code,output,error = self.run_system("/bin/bash -e PATCH/patch.sh",tmp_dir)

            self._check_newdeb_md5sum(tmp_dir+'NEW.file',params['NEW/MD5sum'])

            if new_path:
                shutil.move(tmp_dir+'NEW.file', new_path)
        finally:
            rmtree(tmp_dir)

        return new_path

    def _prepare_env_from_params(self,pkg_cache,params,tmp_dir,olddeb,diversions):
        for a in params:
            if 'needs-old' == a:  # 检查是否需要旧的deb包 可以不需要
                if os.path.isdir(olddeb):
                    logging.error('This patch needs the old version Debian package')
                    raise DebDeltaError(ERROR_UPDATE_DEFAULT_FAILED)
            elif 'old-data-tree' == a:  # 检查是否需要旧包的data 内容
                os.mkdir(tmp_dir+'/OLD/DATA')
                # 从本地进行构建旧包
                if os.path.isdir(olddeb):
                    installed_files = self._get_installed_files(pkg_cache,diversions)
                    self._symlink_data_tree(tmp_dir,installed_files)
                # 提供旧包进行构建
                else:
                    logging.error('Needing use old to build new deb')
            elif 'old-control-tree' == a:  # 检查control 文件进行还原
                if os.path.isdir(olddeb):
                    if not os.path.isdir(tmp_dir+'OLD/CONTROL'):
                        os.mkdir(tmp_dir+'OLD/CONTROL')
                    p = params['OLD/Package']
                    a = params['OLD/Architecture']
                    for b in DPKG_KEEPS_CONTROLS:
                        # 组合/var/lib/dpkg/info/kylin-system-updater:all.config
                        z = '/var/lib/dpkg/info/' + p + ':' + a + '.'+b
                        if not os.path.exists(z):
                            # 组合目录 /var/lib/dpkg/info/kylin-system-updater.conffiles
                            z = '/var/lib/dpkg/info/' + p + '.'+b
                        # 文件存在则直接创建软连接 
                        if os.path.exists(z):
                            os.symlink(z, tmp_dir+'OLD/CONTROL/'+b)
                    del z, p  # cannot delete 'a', python raise a SyntaxError
                # else... we always unpack the control of a .deb
            elif 'needs-xdelta3' == a:  # 需要的一些工具进行检查 帮助合成软件包
                if not os.path.exists('/usr/bin/xdelta3'):
                    logging.error(
                        'This patch needs xdelta3. Please install the Debian package "xdelta3".')
            elif 'needs-xdelta3-fifo' == a:
                # not doing a specific check, I am using debian/control Conflicts
                if not os.path.exists('/usr/bin/xdelta3'):
                    logging.error(
                        'This patch needs xdelta3, at least version 3.0y. Please install the Debian package "xdelta3".')
            elif 'needs-xdelta' == a:
                if not os.path.exists('/usr/bin/xdelta'):
                    logging.error(
                        'This patch needs xdelta. Please install the Debian package "xdelta".')
            elif 'needs-bsdiff' == a:
                if not os.path.exists('/usr/bin/bsdiff'):
                    logging.error(
                        'This patch needs bsdiff. Please install the Debian package "bsdiff".')
            elif 'needs-lzma' == a:
                if not os.path.exists('/usr/bin/lzma'):
                    logging.error(
                        'This patch needs lzma. Please install the Debian package "lzma".')
            elif 'needs-xz' == a:
                if not os.path.exists('/usr/bin/xz'):
                    logging.error(
                        'This patch needs xz. Please install the Debian package "xz-utils".')
            elif 'needs-minibzip2' == a:
                pass  # its your lucky day
            elif a[:6] == 'needs-':
                logging.error(
                    'patch says "'+a+"' and this is unsupported. Get a newer debdelta.")
            elif params[a] == True:
                logging.warning(
                    'WARNING patch says %r and this is unsupported. Get a newer debdelta.', a)

    def _my_popen_read(self,cmd, ** args):
        return subprocess.Popen(cmd, stdout=subprocess.PIPE, stdin=open(os.devnull),
                                close_fds=True, **args).stdout

    def _save_to_local(self,data,output_path):
        with open(output_path, "wb+") as f:
            f.write(data)

    def info_2_db(self,info):
        params = {}
        for s in info:
            if ':' in s:
                i = s.index(':')
                params[s[:i]] = s[i+2:]
            elif s:
                params[s] = True
        return params

    def run_system(self,args,tmp_dir):
        try:
            process = subprocess.Popen(args, shell=True,cwd=tmp_dir,stdout=subprocess.PIPE,stderr=subprocess.PIPE,text=True)
            
            output, error = process.communicate()
            code = process.returncode

            return code,output,error
        except Exception as exc:
            logging.error(exc)

    def _info_patch_unzip(self,tmp_dir):
        "unzip info and patch.sh"
        if os.path.exists(tmp_dir+'PATCH/info.gz'):
            self.run_system("gunzip PATCH/info.gz", tmp_dir)
        if os.path.exists(tmp_dir+'PATCH/patch.sh.gz'):
            self.run_system("gunzip PATCH/patch.sh.gz", tmp_dir)
        elif os.path.exists(tmp_dir+'PATCH/patch.sh.bz2'):
            self.run_system("bunzip2 PATCH/patch.sh.bz2", tmp_dir)
        elif os.path.exists(tmp_dir+'PATCH/patch.sh.lzma'):
            if not os.path.exists('/usr/bin/unlzma'):
                logging.error("This patch needs lzma. Please install the Debian package lzma")
                raise DebDeltaError(ERROR_UPDATE_DEFAULT_FAILED)
            self.run_system("unlzma PATCH/patch.sh.lzma",tmp_dir)
        elif os.path.exists(tmp_dir+'PATCH/patch.sh.xz'):
            if not os.path.exists('/usr/bin/unxz'):
                logging.error("This patch needs xz. Please install the Debian package xz-utils")
                raise DebDeltaError(ERROR_UPDATE_DEFAULT_FAILED)
            self.run_system("unxz PATCH/patch.sh.xz", tmp_dir)
    
    def _get_installed_files(self,pkg_cache,diversions):
        data = []
        for files in pkg_cache.installed_files:
            if files == '':
                continue
            if files in diversions:
                diver_file,dever_name = diversions[files]
                if dever_name != pkg_cache.name :
                    data.append((files,diver_file))
                else:
                    data.append((files,files))
            else:
                data.append((files,files))
        return data
        
    def _symlink_data_tree(self,tmp_dir,installed_files):
        for orig, divert in installed_files:
            if os.path.isfile(divert) and not os.path.islink(divert):
                tmpcopy = tmp_dir+'OLD/DATA'+orig
                d = os.path.dirname(tmpcopy)  # 检查目录是否存在 不存在创建
                if not os.path.exists(d):
                    os.makedirs(d)
                os.symlink(divert, tmpcopy)  # 直接建立软连接
            elif not os.path.exists(divert) and os.path.islink(divert):
                logging.error('  Broken symlink? %r', divert)
            elif not os.path.exists(divert):
                logging.error(' Disappeared file? %r', divert)
                for z in ('locale', 'man', 'gnome/help', 'omf', 'doc/kde/HTML'):
                    w = '/usr/share/'+z
                    if orig[:len(w)] == w:
                        logging.error(' Disappeared localepurged? %s', str(orig))
            else:
                logging.debug('    not symlinking %r to %r', divert, orig)
        return True

    def _scan_delta_info(self,tmp_dir):
        info = []
        if os.path.isfile(tmp_dir+'PATCH/info'):
            # new style debdelta, with info file
            p = open(tmp_dir+'PATCH/info')
            info = p.read().split('\n')
            p.close()
            if info[-1] == '':
                info.pop()
        else:
            # old style debdelta, with info in patch.sh
            p = open(tmp_dir+'PATCH/patch.sh')
            s = p.readline()
            s = p.readline()
            while s:
                if s[0] == '#':
                    s = self._de_n(s)
                    info.append(s[1:])
                s = p.readline()
            p.close()
        return info

    def _check_is_delta(self,f):
        # FIXME:检查出现异常直接抛出
        if not os.path.exists(f):
            logging.error(_("Error: the file '%s' does not exist.") % f)
            raise DebDeltaError(ERROR_UPDATE_DEFAULT_FAILED)
        if not os.path.isfile(f):
            logging.error(_("Error: '%s' is not a regular file.") % f)
            raise DebDeltaError(ERROR_UPDATE_DEFAULT_FAILED)
        p = open(f, 'rb')
        if p.read(8) != b'!<arch>\n':
            logging.error(_("Error: '%s' does not seem to be a Debian delta.") % f)
            raise DebDeltaError(ERROR_UPDATE_DEFAULT_FAILED)
        p.close()

    def _check_deb(self,f):
        # FIXME:检查出现异常直接抛出
        if not os.path.exists(f):
            logging.error(_("Error: the file '%s' does not exist.") % f)
            raise DebDeltaError(ERROR_UPDATE_DEFAULT_FAILED)
        if not os.path.isfile(f):
            logging.error(_("Error: '%s' is not a regular file.") % f)
            raise DebDeltaError(ERROR_UPDATE_DEFAULT_FAILED)
        p = open(f, 'rb')
        if p.read(21) != b"!<arch>\ndebian-binary":
            raise DebDeltaError(ERROR_UPDATE_DEFAULT_FAILED)
        p.close()

    def _calculate_md5(self,file_path):
        md5_hash = hashlib.md5()

        with open(file_path, 'rb') as file:
            for chunk in iter(lambda: file.read(4096), b''):
                md5_hash.update(chunk)

        # 获取计算得到的 MD5 值
        md5_value = md5_hash.hexdigest()
        return md5_value

    def _check_newdeb_md5sum(self,deb_file,targe_value):
        md5_sum = self._calculate_md5(deb_file)

        if targe_value != md5_sum:
            logging.error(' MD5 mismatch, '+str(targe_value)+' != ' + str(md5_sum))
            # raise ("MD5 mismatch")
            return False
        else:
            return True

    def _scan_diversions(self):
        with open('/var/lib/dpkg/diversions', "r") as f:
            d={}
            a=1
            while 1:
                a=f.readline()
                if not a: break
                a=self._de_n(a)
                b=self._de_n(f.readline())
                p=self._de_n(f.readline())
                d[a]=(b,p)
        return d

    def _de_n(self,a):
        assert isinstance(a, str)
        if a and a[-1:] ==  '\n' :
            a = a[:-1]
        return a
    
    def _make_temp_dir(self):
        tmp_dir = abspath(tempfile.mkdtemp(prefix='debdelta',dir=TMPDIR))
        for i in 'OLD','NEW','PATCH' :
            os.mkdir(tmp_dir+'/'+i)
        
        if tmp_dir[-1] != '/':
            tmp_dir = tmp_dir + '/'
        return tmp_dir


if __name__ == "__main__":
    import sys
    cache = Cache(None)
    tt = DeltaDownloadBackend(cache)
    
    FORMAT = "%(asctime)s [%(levelname)s]: %(message)s"
    logging.basicConfig(format=FORMAT,level=logging.INFO,datefmt='%m-%d,%H:%M:%S')

    def _test_download_delta():
        #测试下载delta文件
        download_session = requests.Session()
        download_session.headers.update(HTTP_USER_AGENT.copy())
        uri = "http://debdeltas.debian.net/debian-deltas/pool/main/b/base-files/base-files_11.1+deb11u6_11.1+deb11u7_amd64.debdelta"
        delta_path = "/home/x/base-files_11.1+deb11u6_11.1+deb11u7_amd64.debdelta"
        tt._download_10k_uri(download_session,uri)

    # _test_download_delta()

    def _test_run_system():
    #测试 执行终端命令
        ret,output,error  = tt.run_system("ar xvo "+ "/home/x/base-files_11.1+deb11u6_11.1+deb11u7_amd64.debdelta","/home/x"+'/PATH')
        ar_list_delta=[a[4:] for a in output.split('\n') if a]

    #测试 读取info文件
    def _test_read_delta_info():
        info = tt._scan_delta_info("/home/x/")

    # 测试打patch
    def _test_do_path():
        tt._do_patch("/home/x/test.debdelta","/","/home/x/test.deb",[],cache["kylin-system-updater"])

    tt.commit_download(['apparmor-notify'])