#!/usr/bin/env python
# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-

"""Integration of package managers into UpdateManager"""
# (c) 2005-2009 Canonical, GPL
import logging
import os
from gettext import gettext as _


class InstallBackend():
    ACTION_DEFUALT_STATUS = -1
    ACTION_UPDATE = 0
    ACTION_INSTALL = 1
    ACTION_INSTALL_DEB = 2
    ACTION_CHECK_RESOLVER = 3
    ACTION_DOWNLOADONLY = 4
    ACTION_FIX_BROKEN = 5
    ACTION_REMOVE_PACKAGES = 6
    ACTION_FIX_INCOMPLETE = 7
    ACTION_CLEAN = 8
    ACTION_INSTALL_SHUTDOWN = 9
    ACTION_BACKGROUND_UPGRADE = 10
    ACTION_INSTALL_DEB_RESOLVER = 11
    ACTION_SLEF_UPGRADE = 12

    MODE_DEFAULT_STATUS = -1
    
    MODE_INSTALL_PARTIAL = 0
    MODE_INSTALL_ALL = 1
    MODE_INSTALL_SYSTEM = 2

    MODE_UPDATE_CACHE = 0
    MODE_UPDATE_ALL = 1

    def __init__(self, window_main, action):
        self.action = action

    def start(self):
        if self.action == self.ACTION_UPDATE:
            self.update()

    def update(self):
        """Run a update to refresh the package list"""
        raise NotImplementedError

# try aptdaemon
if os.path.exists("/usr/sbin/aptd") \
   and "UPDATE_MANAGER_FORCE_BACKEND_SYNAPTIC" not in os.environ:
    # check if the gtkwidgets are installed as well
    try:
        from .InstallBackendAptdaemon import InstallBackendAptdaemon
    except ImportError:
        logging.exception("importing aptdaemon")

def get_backend(*args, **kwargs):
    """Select and return a package manager backend."""
    # try aptdaemon
    if (os.path.exists("/usr/sbin/aptd")
            and "UPDATE_MANAGER_FORCE_BACKEND_SYNAPTIC" not in os.environ):
        # check if the gtkwidgets are installed as well
        try:
            return InstallBackendAptdaemon(*args, **kwargs)
        except NameError:
            logging.exception("using aptdaemon failed")

    # nothing found, raise
    raise Exception("No working backend found, please try installing "
                    "aptdaemon or synaptic")